{***********************************************************************
Unit gfx_basedef.PAS v1.5 0801
    (c) by Andreas Moser, AMoser@amoser.de

    Delphi version : Delphi 4, 6 (due to user responses it seem to work with Delphi 5)

    gfx_basedef is part of the gfx_library collection

    You may use this sourcecode for your freewareproducts.
    You may modify this source-code for your own use.
    You may recompile this source-code for your own use.

    All functions, procedures and classes may NOT be used in commercial
    products without the permission of the author. For parts of this library
    not written by me, you have to ask for permission by their
    respective authors.

    Disclaimer of warranty: "This software is supplied as is. The author
    disclaims all warranties, expressed or implied, including, without
    limitation, the warranties of merchantability and of fitness for any
    purpose. The author assumes no liability for damages, direct or
    consequential, which may result from the use of this software."

    All brand and product names are marks or registered marks of their
    respective companies.

    Please report bugs to:
    Andreas Moser  amoser@amoser.de

    UPDATES
    v1.1 04/2001
                 - bugfix gfx_errors (for delphi 5 compatibility)
                 - update function SetBmpEqual
                 - new unit gfx_crypt
                 - new predefined filters in gfx_effectexd (glow, waggle, pattern, churchpattern)
    v1.2 08/2001 - new unit gfx_mask

    v1.5 08/2001 - rework of gfx_effects
                   - functionallity is now in "gfx_DIBProcessing"
                 - New function HSLtoRGB2
                 - New unit "gfx_syswrapper"
                        - all system dependent functions, classes and procedures are now
                          wrapped by this unit.
                        - now only gfx_basedef & gfx_syswrapper needs system units
                          like "windows, graphics, sysutils" etc.
                        - It would be easier for someone to convert gfx_lib to another
                          language or platform
********************************************************************************}
{$define DIB_INTFC}

{.$DEFINE PIXELFORMAT_TOO_SLOW}
{.$DEFINE CREATEDIBSECTION_SLOW}

// Delphi 1.x
{$IFDEF VER80}
  'Error: Library does not support Delphi 1.x'
{$ENDIF}

// Delphi 2.x
{$IFDEF VER90}
  'Error: Library does not support Delphi 2.x'
{$ENDIF}

// C++ Builder 1.x
{$IFDEF VER93}
  'Error: Library does not support C++ Builder 1.x'
{$ENDIF}

// Delphi 3.x
{$IFDEF VER100}
  'Error: Library does not support Delphi 3.x'
{$ENDIF}

// C++ Builder 3.x
{$IFDEF VER110}
  'Error: Library does not support C++ Builder 3.x'
{$ENDIF}

// Delphi 4.x
{$IFDEF VER120}
  {$DEFINE VER10_PLUS}
  {$DEFINE VER11_PLUS}
  {$DEFINE VER12_PLUS}
  {$DEFINE BAD_STACK_ALIGNMENT}
{$ENDIF}

// C++ Builder 4.x
{$IFDEF VER125}
  {$DEFINE VER10_PLUS}
  {$DEFINE VER11_PLUS}
  {$DEFINE VER12_PLUS}
  {$DEFINE VER125_PLUS}
  {$DEFINE BAD_STACK_ALIGNMENT}
{$ENDIF}

// Delphi 5.x
{$IFDEF VER130}
  {$DEFINE VER10_PLUS}
  {$DEFINE VER11_PLUS}
  {$DEFINE VER12_PLUS}
  {$DEFINE VER125_PLUS}
  {$DEFINE VER13_PLUS}
  {$DEFINE BAD_STACK_ALIGNMENT}
{$ENDIF}

// Unknown compiler version - assume D4 compatible
{$IFNDEF VER9x}
  {$IFNDEF VER10_PLUS}
    {$DEFINE VER10_PLUS}
    {$DEFINE VER11_PLUS}
    {$DEFINE VER12_PLUS}
    {$DEFINE BAD_STACK_ALIGNMENT}
  {$ENDIF}
{$ENDIF}

unit gfx_basedef;

interface

uses Windows,Graphics, Messages, Classes, SysUtils, gfx_syswrapper, gfx_dib;

// -----------------------------------------------------------------------------
//
//			Types
//
// -----------------------------------------------------------------------------
type
   {Error handling}
   EGraphicEffects = class(Exception);
   EGraphicConvert = class(Exception);
   EGraphicFormat = class(Exception);



const   { Ranges of the IEEE floating point types, including denormals }
  MaxEffectMatrixVector    =   7;
  MaxEffectMatrixSize      =   49;

  MinSingle        =     1.5e-45;
  MaxSingle        =     3.4e+38;
  MinDouble        =     5.0e-324;
  MaxDouble        =     1.7e+308;
  MinExtended      =     3.4e-4932;
  MaxExtended      =     1.1e+4932;
  MinComp          =     -9.223372036854775807e+18;
  MaxComp          =     9.223372036854775807e+18;

  MaxPixelCount = 32768;

  BitmapAllocationThreshold = 500000;	// Bitmap pixel count limit at which
  					// a newly allocated bitmap will be
                                        // converted to 1 bit format before
                                        // being resized and converted to 8 bit.
  type
    TIntVector     = array[0..100] of Integer;
    PIntVector     = ^TIntVector;

    TIntXY         = record
                  Int            : Integer;
                  X,Y            : Integer;
                  end;

    TIntXYVector   = array [0..MaxEffectMatrixSize] of TIntXY;


    {definiton used by EffectSwapChannels}
    TSwapPlanes= (swapNone,swapRedBlue,swapRedGreen,swapBlueGreen);

    {definition used by EffectArithmetic}
    TGraphicArithmetic=(gaNone,gaAdd,gaSubstract,gaMultiply,gaDivide,gaDarkest,
                        gaLightest,gaDifference,gaBinaryOr,gaBinaryAnd,gaAverage,gaBlend);


    TFXType =(fxAtithmetiv, fxBlend, fxAddNoise, fxEllipse, fxGrayScale, fxOldLook, fxFishEye, fxGamma,
              fxCircleAround, fxSolarize, fxNegative, fxSpray, fxPosterize, fxAntiAlias, fxContrast,
              fxLightNess, fxIncdecRGB, fxColorize, fxMosaic, fxStretch, fxSinus, fxSwapChannels, fxFilter);

    type TDirection      = (TopToBtm, BtmToTop, LftToRgt,RgtToLft,TopLftToBtmRgt, BtmLftToTopRgt, All);

  TColorReduction =
    (rmNone,			// Do not perform color reduction
     rmWindows20,		// Reduce to the Windows 20 color system palette
     rmWindows256,		// Reduce to the Windows 256 color halftone palette (Only works in 256 color display mode)
     rmWindowsGray,		// Reduce to the Windows 4 grayscale colors
     rmMonochrome,		// Reduce to a black/white monochrome palette
     rmGrayScale,		// Reduce to a uniform 256 shade grayscale palette
     rmNetscape,		// Reduce to the Netscape 216 color palette
     rmQuantize,		// Reduce to optimal 2^n color palette
     rmQuantizeWindows,		// Reduce to optimal 256 color windows palette
     rmPalette,			// Reduce to custom palette
     rmStandard,
     rmDesktop
    );
  TDitherMode =
    (dmNone,
     dmNearest,			// Nearest color matching w/o error correction
     dmFloydSteinberg,		// Floyd Steinberg Error Diffusion dithering
     dmStucki,			// Stucki Error Diffusion dithering
     dmSierra,			// Sierra Error Diffusion dithering
     dmJaJuNI,			// Jarvis, Judice & Ninke Error Diffusion dithering
     dmSteveArche,		// Stevenson & Arche Error Diffusion dithering
     dmBurkes			// Burkes Error Diffusion dithering
     // dmOrdered,		// Ordered dither
    );



    {---------------------------------

      Colorspace def

    ----------------------------------}

    TColorSpace = (csRGB, csHSV, csHSL, csCMYK, csRGBA, csBGR, csBGRA, csCMY, csCIELAB, csYCBCR, cs8Bit,cs1Bit,cs4Bit,csNone);
    TColorPlane = (cpRGB,cpRed, cpGreen, cpBlue,
                   cpHueHSV, cpSaturationHSV, cpValue,
                   cpCyan, cpMagenta, cpYellow, cpBlack,
                   cpIntensity,
                   cp8bit);



    {---------------------------------

      Effect Matrices and types

    ----------------------------------}

    {Static filters}
    TStaticFilterType=(sfMedian,sfMin,sfMax);

    {Filter types}
    TFilterType=(ftStatic,ftLinear,ftMultiPass);

    {Matrix size (3x3, 5x5, 7x7}
    TMatrixSize = (mxDefault,mx3,mx5,mx7,mxUndefined);

    {TEffectMatrix used by EffectLinearFilterRGB}
    TEffectMatrix = array[0..MaxEffectMatrixVector-1,0..MaxEffectMatrixVector-1] of Integer;
    pEffectMatrix = ^TEffectMatrix;

    {TGraphicFilter used by EffectLinearFilterRGB /EffectStaticFilterRGB}
    TGraphicFilter = packed record
      FilterType:   TFilterType;
      MatrixSize:   TMatrixSize;     // Size: mx3=3x3, mx5=5x5, mx7=7x7
      Matrix:       TEffectMatrix;   // Every pixel of source would be calculated by the matrix
      Divisor:      Integer;         // The result of calculation would be divided by divisor (>0!)
      Bias:         Integer;         // Finally bias value would be added
      FilterName:   Array [0..128] of Char;
    end;
    pGraphicFilter=^TGraphicFilter;

    {TMultiPassGraphicFilter used by EffectMultiPassFilter}
    TMultiPassGraphicFilter=packed record
       FilterType:TFilterType;
       Filters:array [1..4] of pGraphicFilter;
       Functions:array [1..3] of TGraphicArithmetic;
       FilterName:Array [0..100] of Char;
    end;
    pMultiPassGraphicFilter=^TMultiPassGraphicFilter;


    TEffectCallBack = procedure (const Min,Max,Pos:Integer);

    {Default Filter (NULL-Filter}
    CONST mxZero:TGraphicFilter
      =(FilterType:ftLinear;
      MatrixSize:mx3;
      Matrix:
     (( 0, 0, 0, 0, 0, 0, 0),
      ( 0, 0, 0, 0, 0, 0, 0),
      ( 0, 0, 0, 0, 0, 0, 0),
      ( 0, 0, 0, 1, 0, 0, 0),
      ( 0, 0, 0, 0, 0, 0, 0),
      ( 0, 0, 0, 0, 0, 0, 0),
      ( 0, 0, 0, 0, 0, 0, 0));
      Divisor:1;
      Bias:0;
      FilterName:'Null (Linear)');

      {Default static filters}
      mxMedian:TGraphicFilter
          =(FilterType:ftStatic;
          MatrixSize:mx3;
          FilterName:'Median (Static)');
      mxMaximum:TGraphicFilter
          =(FilterType:ftStatic;
          MatrixSize:mx3;
          FilterName:'Maximum (Static)');
      mxMinimum:TGraphicFilter
          =(FilterType:ftStatic;
          MatrixSize:mx3;
          FilterName:'Minimum (Static)');

    {---------------------------------

      Palette types

    ----------------------------------}
    Type
    TPaletteEntries = array[BYTE] of TPaletteEntry;
    pPaletteEntries = ^TPaletteEntries;
    TPaletteArray = Array [0..2,0..255] of Byte;

    TPaletteWordArray = Array [0..2,0..255] of Word;

    const
    PaletteVersion = $0300;
    clMoneyGreen = TColor($C0DCC0);   // Color   "8"  RGB:  192 220 192
    clSkyBlue    = TColor($F0CAA6);   // Color   "9"  RGB:  166 202 240
    clCream      = TColor($F0FBFF);   // Color "246"  RGB:  255 251 240
    clMediumGray = TColor($A4A0A0);   // Color "247"  RGB:  160 160 164

    {---------------------------------

      Colorspace types

    ----------------------------------}
    type
    TColorRGB = packed record
    r, g, b	: BYTE;
    end;
    PColorRGB = ^TColorRGB;

    pRGBarray  = ^TRGBarray;
    TRGBarray   = array[0..MaxPixelCount-1] of TRGBTriple;
    TPRows = array[0..1000000] of pRGBarray;
    PPRows = ^TPRows;

    pRGBTriple = ^TRGBTriple;
    TRGBQuadArray   = array[BYTE] of TRGBQUAD;

    TRGBTripleArray = array of TRGBTriple;

    TScanLine= Array of Byte;
    pScanLine= ^TScanLine;

    TBytearrayL = array[0..65535] of Byte;
    pBytearrayL = ^TBytearrayL;

    T3Bytearray = array[0..3*65535] of Byte;
    p3Bytearray = ^TBytearray;

    TCardinalArray = Array of Cardinal;

    TCMYKQuad = record
       cmykCyan,cmykMagenta,cmykYellow,cmykK: Byte;
    end;
    pCMYKQuad = ^TCMYKQuad;

    TCMYTriple = record
       cmykCyan,cmykMagenta,cymkYellow: Integer;
    end;
    pCMYTriple = ^TCMYTriple;

    THSVTriple = record
       hsvHue,hsvSaturation,hsvValue : Integer;
    end;
    pHSVTriple = ^THSVTriple;

    THSLTriple = record
       hslHue,hslSaturation,hslLightness : Integer;
    end;
    pHSLTriple = ^THSLTriple;

    TRGBAQuad = record
       rgbBlue,rgbGreen,rgbRed,rgbAlpha: Integer;
    end;
    pRGBAQuad = ^TRGBAQuad;

    TBGRTriple = record
       bgrBlue,bgrGreen,bgrRed: Integer;
    end;
    pBGRTriple = ^TBGRTriple;

    TBGRAQuad = record
       bgrBlue,bgrGreen,bgrRed,bgrAlpha: Integer;
    end;
    pBGRAQuad = ^TBGRAQuad;

    TBGRTripleByte = record
       bgrBlue,bgrGreen,bgrRed: Byte;
    end;
    pBGRTripleByte = ^TBGRTripleByte;

    TBGRAQuadByte = record
       bgrBlue,bgrGreen,bgrRed,bgrAlpha: Byte;
    end;
    pBGRAQuadByte = ^TBGRAQuadByte;

    TYCbCrTriple = record
      ycbcrY,ycbcrCB,ycbcrCR:Integer;
    end;
    pYCbCrTriple = ^TYCbCrTriple;

    TCIELABTriple = record
      L,A,B:Byte;
    end;
    pCIELABTriple = ^TCIELABTriple;

    TBool2dArray=array of array of boolean;

//var ScanLineArray:Array[0..65535] Of Byte;
    type
    TBMPInterface= class (TObject)
    private
      fSourceBitmap   : TBitmap;
      {$ifdef DIB_INTFC}
      fSourceDIB      : TDIBReader;
      {$endif}
      fSourceRows     : PPRows;
      fWidth          : Integer;
      fHeight         : Integer;
    protected
      function GetSourceScanLine(Row: integer): pointer;
      function GetSourceRGBTriple(Row,Col: integer): TRGBTriple;
    public
      constructor Create(SourceBitmap:TBitmap);
      destructor  Destroy;override;
      {$ifdef DIB_INTFC}
      property SourceDIB:TDIBReader read fSourceDIB;
      {$endif}
      property SourceBitmap    :TBitmap read fSourceBitmap;
      property SourceScanline[Row:Integer]:Pointer read GetSourceScanline;
      property SourceRGBTriple[Row,Col:Integer]:TRGBTriple read GetSourceRGBTriple;
      property Height:Integer read fHeight;
      property Width :Integer read fWidth;
    end;

    TBMPInterface1= class (TBMPInterface)
    private
      fDestBitmap     : TBitmap;
      fDestPixelFormat: TPixelFormat;
      fDestPalette    : HPalette;
      {$ifdef DIB_INTFC}
      fDestDIB        : TDIBWriter;
      {$endif}
    protected
      function GetDestScanLine(Row: integer): pointer;
      procedure SetDestPalette(Value:HPalette);
    public
      constructor Create(SourceBitmap, DestBitmap:TBitmap;DestPixelFormat:TPixelFormat;DestPalette:HPalette;SameDimensions:Boolean);
      destructor  Destroy;override;
      {$ifdef DIB_INTFC}
      property DestDIB:TDIBWriter read fDestDIB;
      {$endif}
      property DestPalette     :HPalette read fDestPalette write SetDestPalette;
      property DestBitmap      :TBitmap read fDestBitmap;
      property DestScanline[Row:Integer]  :Pointer read GetDestScanline;
    end;


    type
    TBMPInterface2 = class (TBMPInterface1)
    private
      fSourceBitmap2   : TBitmap;
      fSourceDIB2      : TDIBReader;
      fSourceRows2     : PPRows;
      function GetSourceScanLine2(Row: integer): pointer;
      function GetSourceRGBTriple2(Row,Col: integer): TRGBTriple;
    public
      constructor Create(SourceBitmap, SourceBitmap2, DestBitmap:TBitmap;DestPixelFormat:TPixelFormat;DestPalette:HPalette;SameDimensions:Boolean);
      destructor  Destroy;override;
      property SourceBitmap2    :TBitmap read fSourceBitmap2;
      property SourceScanline2[Row:Integer]:Pointer read GetSourceScanline2;
      property SourceRGBTriple2[Row,Col:Integer]:TRGBTriple read GetSourceRGBTriple2;
    end;



    {---------------------------------

      function & procedure interface

    ----------------------------------}

    {because that borland has not included the math.pas to all delphi versions
     here are some math functions}

    function xArcCos(X: Extended): Extended;
    function xArcSin(X: Extended): Extended;
    function xArcTan2(Y, X: Extended): Extended;
    function xCeil(X: Extended): Integer;
    function xFloor(X: Extended): Integer;

    { some min-max functions }
    function xMinValue(const Data: array of Double): Double;
    function xMinIntValue(const Data: array of Integer): Integer;
    function xMaxValue(const Data: array of Double): Double;
    function xMaxIntValue(const Data: array of Integer): Integer;
    function xMulDiv(nNumber,nNumerator,nDenominator:Extended):Integer;

    function TrimReal(Lower, Upper: integer; x: double): integer;     //Trim the real value X into the bounds Lower and Upper
    function TrimInt(Lower, Upper, X: integer): integer;              //Trim the integer value X into the bounds Lower and Upper
    function LoopInt(Lower, Upper, X: integer): integer;              //Trim the integer value X into the bounds Lower and Upper, (if X > Upper then X= Lower +(X-Upper)
    function LoopReal(Lower, Upper:Integer; X: Double): integer;      //Trim the integer value X into the bounds Lower and Upper, (if X > Upper then X= Lower +(X-Upper)
    function MinInt2(Value1,Value2:Integer):Integer;                  //Get the maximum of two integers
    function MaxInt2(Value1,Value2:Integer):Integer;                  //Get the minimum of two integers
    function MinDouble2(Value1,Value2:Double):Double;                 //Get the maximum of two double values
    function MaxDouble2(Value1,Value2:Double):Double;                 //Get the minimum of two double values
    procedure MinMaxInt3(const Value1,Value2,Value3:  INTEGER; var min, max:  INTEGER); //Get the min/max of 3 integers
    procedure MinMaxDouble3(const Value1,Value2,Value3:  Double; var min, max:  Double);//Get the min/max of 3 double values
    procedure SetBitmapsEql(const SrcBitmap,DstBitmap:TBitmap);       //Sets the dimensions of DestBitmap to the values of SrcBitmap, an set the pixelformat of both bitmaps to pf24bit
    function CheckPixelFormat(const Bitmap:TBitmap;RequiredPixelFormat:TPixelFormat):Boolean;
    procedure CopyBitmap(const SrcBitmap,DstBitmap:TBitmap);

    { Color Conversions }
    function RGBTriple(const red, green, blue:  BYTE):  TRGBTriple;
    function RGBtoColor(RGBTriple: TRGBTriple): TColor;
    function ColorToRGB(Color: TColor): TRGBTriple;
    procedure AssignRGBTriple (var TargetTriple:  TRGBTriple;const OriginTriple:  TRGBTriple);
    function RGBIntensity(const RGB:  TRGBTriple):  INTEGER;
    function RGBLightness(const RGB:  TRGBTriple):  INTEGER;
    function RGBSaturation(const RGB:  TRGBTriple):  INTEGER;
    function RGBValue(const RGB:  TRGBTriple):  INTEGER;
    procedure RGBtoCMY(const RGB:  TRGBTriple;var CMY:  TCMYTriple);
    procedure RGBtoCMYK(const RGB:  TRGBTriple;var CMYK:  TCMYKQuad);
    procedure CorrectCMYKColors(var CMYK:TCMYKQuad);
    procedure CMYKtoRGB(const CMYK:  TCMYKQuad; var RGB:TRGBTriple);
    procedure HSVtoRGB (const HSV: THSVTriple ;var   RGB:TRGBTriple);  // S, V must be in [0..255], H must be in [0..359]
    procedure RGBtoHSV (const RGB:TRGBTriple;var HSV:THSVTriple);
    procedure RGBtoHSLDbl (RGB: TRGBTriple; var H,S,L:Double);
    procedure RGBtoHSL (RGB: TRGBTriple; var HSL:THSLTriple);
    procedure HSLDblToRGB (H,S,L:Double;var RGB:  TRGBTRiple);        // S, L must be in [0..255], H must be in [0..359]
    procedure HSLtoRGB (HSL:THSLTriple;var RGB:  TRGBTRiple);          // S, L must be in [0..255], H must be in [0..359]
    procedure HSLtoRGB2 (HSL:THSLTriple;var R,G,B:  Byte);          // S, L must be in [0..255], H must be in [0..359]
    procedure YCbCrToRGB(ycbcr:TYCbCrTriple; var RGB:  TRGBTRiple);
    procedure CIELABtoBGR(CIELABTriple:TCIELABTriple; var BGRTriple:TBGRTriple);
    procedure RGBAtoBGRA(RGBA:TRGBAQuad; var BGRA:TBGRAQuad);
    procedure RGBAtoBGRA_16(RGBA:TRGBAQuad; var BGRA:TBGRAQuad);

    {Gamma}
    function CalculateGamma(Gamma:Double;Value:Integer):Double;
    function IntGamma(Gamma:Double;Value:Integer):Integer;

    {DC Info}
    function  GetColorDepth:  INTEGER;
    function  GetPixelDepth:  INTEGER;
    function  IsPaletteDevice:  BOOLEAN;

    {Bitmaps info & utility}
    procedure PrintBitmap(Canvas:  TXCanvas; DestRect:  TXRect;  Bitmap:  TXBitmap);
    function GetBitmapDimensionsString(const Bitmap:  TXBitmap):  string;
    function GetColorPlaneString(const ColorPlane:  TColorPlane):  string;
    function GetPixelformatString(const Pixelformat:  TXPixelformat):  string;


    {Filter helpers}
    procedure RotateFilterMatrix(var Filter:TGraphicFilter;Steps:Integer);
    function GetStatOfMatrix(Matrix:TEffectMatrix;SFType:TSTaticFilterType;Diameter:Integer;var Row,Col:Integer):Integer;    procedure QSort(var X : TIntVector; Lbound, Ubound : Integer);
    procedure MatrixQSort(var X : TIntXYVector;Size:Integer);
    procedure SaveLinearFilterToFile(FileName:PChar;const Filter:TGraphicFilter);stdcall;
    procedure LoadLinearFilterFromFile(FileName:PChar;var Filter:TGraphicFilter);stdcall;
    procedure SaveMultipassFilterToFile(FileName:PChar;const Filter:TMultiPassGraphicFilter);stdcall;
    procedure LoadMultiPassFilterFromFile(FileName:PChar;var Filter:TMultiPassGraphicFilter);stdcall;

{ ---------------------------------------------------------------- }

implementation
uses gfx_errors;

// -----------------------------------------------------------------------------
//
//			TBMPInterface
//
// -----------------------------------------------------------------------------
constructor TBMPInterface.Create(SourceBitmap:TBitmap);
var i:Integer;
begin

  inherited Create;
  fSourceBitmap:=SourceBitmap;
  fHeight:=fSourceBitmap.Height;
  fWidth :=fSourceBitmap.Width;
  {$ifdef DIB_INTFC}
  fSourceDIB := TDIBReader.Create(fSourceBitmap, fSourceBitmap.PixelFormat);
  {$endif}
  GetMem(fSourceRows, fSourceBitmap.Height * SizeOf(pRGBArray));
  {$ifdef DIB_INTFC}
  for i:= 0 to fSourceBitmap.Height - 1 do fSourceRows[i]:=fSourceDIB.ScanLine[i];
  {$else}
  for i:= 0 to fSourceBitmap.Height - 1 do fSourceRows[i]:=fSourceBitmap.Scanline[i];
  {$endif}
end;

destructor TBMPInterface.Destroy;
begin
  {$ifdef DIB_INTFC}
   if (fSourceDIB <> nil) then fSourceDIB.Free;
  {$endif}
   FreeMem(fSourceRows);
  inherited Destroy;
end;

function TBMPInterface.GetSourceScanLine(Row: integer): pointer;
begin
  {$ifdef DIB_INTFC}
  Result := fSourceDIB.ScanLine[Row];
  {$else}
  Result := FSourceBitmap.ScanLine[Row];
  {$endif}
end;

function TBMPInterface.GetSourceRGBTriple(Row,Col: integer): TRGBTriple;
begin
 if (Row>=0) and (Row <fSourceBitmap.Height) and (Col>=0) and (Col<fSourceBitmap.Width) then
  result :=fSourceRows[Row][Col] else
  with Result do
  begin
    rgbtBlue:=0;
    rgbtGreen:=0;
    rgbtRed:=0;
  end;
end;

// -----------------------------------------------------------------------------
//
//			TBMPInterface1
//
// -----------------------------------------------------------------------------
constructor TBMPInterface1.Create(SourceBitmap, DestBitmap:TBitmap; DestPixelFormat:TPixelFormat;DestPalette:HPalette;SameDimensions:Boolean);
begin
  inherited Create(SourceBitmap);
  fDestBitmap:=DestBitmap;
  fdestPixelFormat:=DestPixelFormat;
  case fDestPixelFormat of
  pf15bit,pf16bit,pf24bit,pf32bit: fDestPixelFormat:=pf24bit;
  else
    fDestPixelFormat:=pf8bit;
  end;
  if SameDimensions then SetBitmapsEql(fSourceBitmap,fDestBitmap);

  {$ifdef DIB_INTFC}
  fDestDIB   := TDIBWriter.Create(fDestBitmap, fDestPixelFormat, fDestBitmap.Width, fDestBitmap.Height,DestPalette);
  {$endif}
end;

destructor TBMPInterface1.Destroy;
begin
  {$ifdef DIB_INTFC}
   if (fDestDIB <> nil) then fDestDIB.Free;
  {$endif}

  inherited Destroy;
end;

function TBMPInterface1.GetDestScanLine(Row: integer): pointer;
begin
  {$ifdef DIB_INTFC}
  Result := fDestDIB.ScanLine[Row];
  {$else}
  Result := FDestBitmap.ScanLine[Row];
  {$endif}
end;

procedure TBMPInterface1.SetDestPalette(Value:HPalette);
begin
  if Value<>fDestPalette then
  begin
    FDestPalette := CopyPalette(Value);
  {$ifdef DIB_INTFC}
    fDestDIB.Bitmap.Palette := FDestPalette;
  {$endif}
  end;
end;

// -----------------------------------------------------------------------------
//
//			TBMPInterface2
//
// -----------------------------------------------------------------------------
constructor TBMPInterface2.Create(SourceBitmap, SourceBitmap2, DestBitmap:TBitmap; DestPixelFormat:TPixelFormat;DestPalette:HPalette;SameDimensions:Boolean);
var i:Integer;
begin
  inherited Create(SourceBitmap,DestBitmap,DestPixelFormat,0,SameDimensions);
  fSourceBitmap2:=SourceBitmap2;
  if SameDimensions then SetBitmapsEql(fSourceBitmap,fSourceBitmap2);
  {$ifdef DIB_INTFC}
  fSourceDIB2 := TDIBReader.Create(fSourceBitmap2, fSourceBitmap2.PixelFormat);
  {$endif}
  GetMem(fSourceRows2, fSourceBitmap2.Height * SizeOf(pRGBArray));
  {$ifdef DIB_INTFC}
  for i:= 0 to fSourceBitmap2.Height - 1 do fSourceRows2[i]:=fSourceDIB2.ScanLine[i];
  {$else}
  for i:= 0 to fSourceBitmap2.Height - 1 do fSourceRows2[i]:=fSourceBitmap2.Scanline[i];
  {$endif}
end;

destructor TBMPInterface2.Destroy;
begin
  {$ifdef DIB_INTFC}
   if (fSourceDIB2 <> nil) then fSourceDIB2.Free;
  {$endif}
   FreeMem(fSourceRows2);

  inherited Destroy;
end;

function TBMPInterface2.GetSourceScanLine2(Row: integer): pointer;
begin
  {$ifdef DIB_INTFC}
  Result := fSourceDIB2.ScanLine[Row];
  {$else}
  Result := FSourceBitmap2.ScanLine[Row];
  {$endif}
end;


function TBMPInterface2.GetSourceRGBTriple2(Row,Col: integer): TRGBTriple;
begin
 if (Row>0) and (Row <=fSourceBitmap2.Height) and (Col>0) and (Col<=fSourceBitmap2.Width) then
  result :=fSourceRows2[Row][Col] else
  with Result do
  begin
    rgbtBlue:=0;
    rgbtGreen:=0;
    rgbtRed:=0;
  end;
end;

// -----------------------------------------------------------------------------
//
//			xArcCos
//
// -----------------------------------------------------------------------------
function xArcCos(X: Extended): Extended;
begin
  Result := xArcTan2(Sqrt(1 - X*X), X);
end;

// -----------------------------------------------------------------------------
//
//			xArcSin
//
// -----------------------------------------------------------------------------
function xArcSin(X: Extended): Extended;
begin
  Result := xArcTan2(X, Sqrt(1 - X*X))
end;

// -----------------------------------------------------------------------------
//
//			xArcTan2
//
// -----------------------------------------------------------------------------
function xArcTan2(Y, X: Extended): Extended;
asm
        FLD     Y
        FLD     X
        FPATAN
        FWAIT
end;

// -----------------------------------------------------------------------------
//
//			xCeil
//
// -----------------------------------------------------------------------------
function xCeil(X: Extended): Integer;
begin
  Result := Trunc(X);
  if Frac(X) > 0 then
    Inc(Result);
end;

// -----------------------------------------------------------------------------
//
//			xFloor
//
// -----------------------------------------------------------------------------
function xFloor(X: Extended): Integer;
begin
  Result := Trunc(X);
  if Frac(X) < 0 then
    Dec(Result);
end;

// -----------------------------------------------------------------------------
//
//			xMinValue
//
// -----------------------------------------------------------------------------
function xMinValue(const Data: array of Double): Double;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result > Data[I] then
      Result := Data[I];
end;

// -----------------------------------------------------------------------------
//
//			xMinIntValue
//
// -----------------------------------------------------------------------------
function xMinIntValue(const Data: array of Integer): Integer;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result > Data[I] then
      Result := Data[I];
end;

// -----------------------------------------------------------------------------
//
//			xMaxValue
//
// -----------------------------------------------------------------------------
function xMaxValue(const Data: array of Double): Double;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result < Data[I] then
      Result := Data[I];
end;

// -----------------------------------------------------------------------------
//
//			xMulDiv
//
// -----------------------------------------------------------------------------
function xMulDiv(nNumber,nNumerator,nDenominator:Extended):Integer;
asm
   fld nNumber
   fld nNumerator
   fmul
   fld nDenominator
   fdiv
//     Result:=MulDiv(nNumber,nNumerator,nDenominator);
end;


// -----------------------------------------------------------------------------
//
//			xMaxIntValue
//
// -----------------------------------------------------------------------------
function xMaxIntValue(const Data: array of Integer): Integer;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result < Data[I] then
      Result := Data[I];
end;


// -----------------------------------------------------------------------------
//
//			TrimReal
//
// -----------------------------------------------------------------------------
function TrimReal(Lower, Upper: integer; x: double): integer;
begin
 if (x < upper) and (x >= lower) then
  result:= trunc(x)
 else
  if x >= Upper then
   result:= Upper
    else
     result:= Lower;
end;

// -----------------------------------------------------------------------------
//
//			TrimReal
//
// -----------------------------------------------------------------------------
function TrimInt(Lower, Upper, X: integer): integer;
begin
 if (X <= Upper) and (X >= Lower) then
  TrimInt:= X
 else
  if X > Upper then
   TrimInt:= Upper
    else
     TrimInt:= Lower;
end;

// -----------------------------------------------------------------------------
//
//			LoopInt
//
// -----------------------------------------------------------------------------
function LoopInt(Lower, Upper, X: integer): integer;
begin
 if (X <= Upper) and (X >= Lower) then
  LoopInt:= X
 else
  if X > Upper then
   LoopInt:= Lower + X-Upper
    else
     LoopInt:= Upper-Lower+X;
end;

// -----------------------------------------------------------------------------
//
//			LoopReal
//
// -----------------------------------------------------------------------------
function LoopReal(Lower, Upper:Integer; X: Double): integer;
begin
 if (X <= Upper) and (X >= Lower) then
  result:= Round(X)
 else
  if X > Upper then
   result:= Round(Lower + X-Upper)
    else
     result:= Round(Upper-Lower+X);
end;

// -----------------------------------------------------------------------------
//
//			MaxInt2
//
// -----------------------------------------------------------------------------
function MaxInt2(Value1,Value2:Integer):Integer;
begin
  if Value1>Value2 then Result:=Value1 else Result:=Value2;
end;

// -----------------------------------------------------------------------------
//
//			MinInt2
//
// -----------------------------------------------------------------------------
function MinInt2(Value1,Value2:Integer):Integer;
begin
  if Value1<Value2 then Result:=Value1 else Result:=Value2;
end;

// -----------------------------------------------------------------------------
//
//			MaxDouble2
//
// -----------------------------------------------------------------------------
function MaxDouble2(Value1, Value2 : double): double;
begin
    if Value1 > Value2 then
      Result := Value1 else  Result := Value2
end;

// -----------------------------------------------------------------------------
//
//			MinDouble2
//
// -----------------------------------------------------------------------------
function MinDouble2 (Value1, Value2 : double): double;
begin
    if Value1 < Value2 then
    Result := Value1 else Result := Value2
end;

// -----------------------------------------------------------------------------
//
//			MinMaxInt3
//
// -----------------------------------------------------------------------------
procedure MinMaxInt3(const Value1,Value2,Value3:  INTEGER; var min, max:  INTEGER);
begin
    if   Value1 > Value2
    then begin
      if   Value1 > Value3
      then max := Value1
      else max := Value3;

      if   Value2 < Value3
      then min := Value2
      else min := Value3
    end
    else begin
      if   Value2 > Value3
      then max := Value2
      else max := Value3;

      if   Value1 < Value3
      then min := Value1
      else min := Value3
    end
end;

// -----------------------------------------------------------------------------
//
//			MinMaxDouble3
//
// -----------------------------------------------------------------------------
procedure MinMaxDouble3(const Value1,Value2,Value3:  Double; var min, max:  Double);
begin
    if   Value1 > Value2
    then begin
      if   Value1 > Value3
      then max := Value1
      else max := Value3;

      if   Value2 < Value3
      then min := Value2
      else min := Value3
    end
    else begin
      if   Value2 > Value3
      then max := Value2
      else max := Value3;

      if   Value1 < Value3
      then min := Value1
      else min := Value3
    end
end;



// -----------------------------------------------------------------------------
//
//			SetBitmapsEql
//
// -----------------------------------------------------------------------------
procedure SetBitmapsEql(const SrcBitmap,DstBitmap:TBitmap);
begin
    DstBitmap.PixelFormat:=SrcBitmap.PixelFormat;
    DstBitmap.Width:=SrcBitmap.Width;
    DstBitmap.Height:=SrcBitmap.Height;
    if SrcBitmap.PixelFormat in [pf1bit, pf4bit, pf8bit]
    then DstBitmap.Palette := CopyPalette(SrcBitmap.Palette);
end;

// -----------------------------------------------------------------------------
//
//			Check PixelFormat
//
// -----------------------------------------------------------------------------
function CheckPixelFormat(const Bitmap:TBitmap;RequiredPixelFormat:TPixelFormat):Boolean;
begin
   if Bitmap.PixelFormat <> RequiredPixelFormat then
   begin
     Error('Invalid Pixelformat');
     Result:=False;
   end else Result:=True;
end;

// -----------------------------------------------------------------------------
//
//			CopyBitmap
//
// -----------------------------------------------------------------------------
procedure CopyBitmap(const SrcBitmap,DstBitmap:TBitmap);
begin
  SetBitmapsEql(SrcBitmap,DstBitmap);
  DstBitmap.Canvas.CopyMode:=cmSrcCopy;
  DstBitmap.Canvas.Draw(0,0,SrcBitmap);

end;


// -----------------------------------------------------------------------------
//
//			RGBTriple
//
// -----------------------------------------------------------------------------
function RGBTriple(const red, green, blue:  BYTE):  TRGBTriple;
begin
  with RESULT DO
  begin
    rgbtRed   := red;
    rgbtGreen := green;
    rgbtBlue  := blue
  end
end;

// -----------------------------------------------------------------------------
//
//			ColorToRGB
//
// -----------------------------------------------------------------------------
function ColorToRGB(Color: TColor): TRGBTriple;
begin
  with Result DO
  begin
    rgbtred := Color AND $000000FF;
    rgbtgreen := (Color AND $0000FF00) SHR 8;
    rgbtblue := (Color AND $00FF0000) SHR 16;
  end;
end;

// -----------------------------------------------------------------------------
//
//			RGBToColor
//
// -----------------------------------------------------------------------------
function RGBtoColor(RGBTriple: TRGBTriple): TColor;
begin
    Result := RGBTriple.rgbtred OR (RGBTriple.rgbtgreen SHL 8) OR (RGBTriple.rgbtblue SHL 16);
end;

// -----------------------------------------------------------------------------
//
//			AssignRGBTriple
//
// -----------------------------------------------------------------------------
procedure AssignRGBTriple (var TargetTriple:  TRGBTriple;const OriginTriple:  TRGBTriple);
begin
    TargetTriple.rgbtRed   := OriginTriple.rgbtRed;
    TargetTriple.rgbtGreen := OriginTriple.rgbtGreen;
    TargetTriple.rgbtBlue  := OriginTriple.rgbtBlue
end {AssignRGBTriple};


// -----------------------------------------------------------------------------
//
//			HslDblToRGB
//
// -----------------------------------------------------------------------------
procedure HSLDblToRGB (H,S,L:Double; var RGB:TRGBTriple);
var
  M1,
  M2: double;

  function HueToColorValue (Hue: double) : byte;
  var
    V : double;
  begin
    if Hue < 0 then
      Hue := Hue + 1
    else

      if Hue > 1 then
        Hue := Hue - 1;

    if 6 * Hue < 1 then
      V := M1 + (M2 - M1) * Hue * 6
    else
      if 2 * Hue < 1 then
        V := M2
      else
        if 3 * Hue < 2 then
          V := M1 + (M2 - M1) * (2/3 - Hue) * 6
        else
          V := M1;
    Result := round (255 * V)
  end;

var
  R,
  G,
  B: byte;
begin
  if S = 0 then
  begin
    R := round (255 * L);
    G := R;
    B := R
  end else begin
    if L <= 0.5 then

      M2 := L * (1 + S)
    else
      M2 := L + S - L * S;
    M1 := 2 * L - M2;
    R := HueToColorValue (H + 1/3);
    G := HueToColorValue (H);
    B := HueToColorValue (H - 1/3)
  end;
  with RGB do
  begin
    rgbtRed:=R;
    rgbtGreen:=G;
    rgbtBlue:=B;
  end;
end;

// -----------------------------------------------------------------------------
//
//			HSLtRGB
//
// -----------------------------------------------------------------------------
procedure HSLToRGB (HSL : THSLTriple;var RGB:TRGBTriple);
begin
  HSLDblToRGB (HSL.hslHue / (359-1), HSL.hslSaturation / 255, HSL.hslLightness / 255,RGB)
end;

// -----------------------------------------------------------------------------
//
//			HSLtRGB2
//
// -----------------------------------------------------------------------------
procedure HSLToRGB2 (HSL : THSLTriple;var R,G,B:Byte);
var RGB:TRGBTriple;
begin
 HSLtoRGB(HSL,RGB);
 R:=RGB.rgbtRed;
 G:=RGB.rgbtGreen;
 B:=RGB.rgbtBlue;
end;

// -----------------------------------------------------------------------------
//
//			HSVtoRGB
//
// -----------------------------------------------------------------------------
procedure HSVtoRGB (const HSV:  THSVTriple;var   RGB  :  TRGBTriple);
var
  fraction :  INTEGER;
  Hue      :  INTEGER;
  p,q,t    :  INTEGER;
  VSSum    :  INTEGER;
  const
      divisor:  INTEGER = 255*60;
begin
  if   HSV.hsvSaturation = 0
  then AssignRGBTriple(RGB, RGBTriple(HSV.hsvValue, HSV.hsvValue, HSV.hsvValue)) else
  begin
    if   HSV.hsvHue = 360 then Hue := 0 else Hue := HSV.hsvHue;
    fraction := Hue MOD 60;
    Hue := Hue div 60;

    VSSum := HSV.hsvValue *HSV.hsvSaturation;
    p := HSV.hsvValue - VSSum div 255;
    q := HSV.hsvValue - (VSSum*fraction) div divisor;
    t := HSV.hsvValue - (VSSum*(60 - fraction)) div divisor;

    with RGB DO
    begin
      CASE Hue of
        0:  begin  rgbtRed := HSV.hsvValue;   rgbtGreen := t;   rgbtBlue := p  end;
        1:  begin  rgbtRed := q;   rgbtGreen := HSV.hsvValue;   rgbtBlue := p  end;
        2:  begin  rgbtRed := p;   rgbtGreen := HSV.hsvValue;   rgbtBlue := t  end;
        3:  begin  rgbtRed := p;   rgbtGreen := q;   rgbtBlue := HSV.hsvValue  end;
        4:  begin  rgbtRed := t;   rgbtGreen := p;   rgbtBlue := HSV.hsvValue  end;
        5:  begin  rgbtRed := HSV.hsvValue;   rgbtGreen := p;   rgbtBlue := q  end;
      end
    end
  end
end;


// -----------------------------------------------------------------------------
//
//			RGBIntensity
//
// -----------------------------------------------------------------------------
{See [Russ95, p. 41]}
function RGBIntensity(const RGB:  TRGBTriple):  INTEGER;
begin
  with RGB DO
    RESULT := (rgbtRed + rgbtGreen + rgbtBlue) div 3
end;


// -----------------------------------------------------------------------------
//
//			RGBLightness
//
// -----------------------------------------------------------------------------
{See [Foley96, p. 595]}
function RGBLightness(const RGB:  TRGBTriple):  INTEGER;
var
    min:  INTEGER;
      max:  INTEGER;
begin
  with RGB DO
    MinMaxInt3(rgbtRed, rgbtGreen, rgbtBlue, min, max);
  RESULT := (min + max) div 2
end ;


// -----------------------------------------------------------------------------
//
//			RGBSaturation
//
// -----------------------------------------------------------------------------
{See [Foley96, p. 592]}
function RGBSaturation(const RGB:  TRGBTriple):  INTEGER;
var
      MaxValue:  INTEGER;
      MinValue:  INTEGER;
begin
     with RGB DO
       MinMaxInt3(rgbtRed, rgbtGReen, rgbtBlue, MinValue, MaxValue);

    {Calculate saturation:  saturation is 0 if r, g and b are all 0}
    if   MaxValue = 0
    then RESULT := 0
    else RESULT := (255 * (MaxValue - MinValue)) div MaxValue;
end;


// -----------------------------------------------------------------------------
//
//			RGBValue
//
// -----------------------------------------------------------------------------
{See [Foley96, p. 592]}
function RGBValue(const RGB:  TRGBTriple):  INTEGER;
begin
    with RGB DO
      RESULT := xMaxIntValue( [rgbtRed, rgbtGreen, rgbtBlue] )
end;


// -----------------------------------------------------------------------------
//
//			RGBtoCMY
//
// -----------------------------------------------------------------------------
// Foley et al, 1996, p. 588
procedure RGBtoCMY(const RGB:  TRGBTriple;    // r, g and b IN [0..255]
                     var   CMY:  TCMYTriple);
begin
    with RGB DO
    begin
      CMY.cmykCyan := 255 - rgbtRed;
      CMY.cmykMagenta := 255 - rgbtGreen;
      CMY.cymkYellow := 255 - rgbtBlue
    end
end;


// -----------------------------------------------------------------------------
//
//			RGBtoCMYK
//
// -----------------------------------------------------------------------------
  // Foley et al, 1996, p. 589
procedure RGBtoCMYK(const RGB:  TRGBTriple;
                      var  CMYK:  TCMYKQuad);
//var cmy:TCMYTriple;
begin
{    //cmy.cmykCyan:=CMYK.cmykCyan;
    //cmy.cmykMagenta:=CMYK.cmykMagenta;
    //cmy.cymkYellow:=CMYK.cmykYellow;
    RGBtoCMY(RGB, CMY);
    CMYK.cmykK := xMinIntValue([CMY.cmykCyan, CMY.cmykMagenta, CMY.cymkYellow]);
    CMYK.cmykCyan := CMY.cmykCyan - CMYK.cmykK;
    CMYK.cmykMagenta := CMY.cmykMagenta - CMYK.cmykK;
    CMYK.cmykYellow := CMY.cymkYellow - CMYK.cmykK
    }

  CMYK.cmykCyan := 255 - RGB.rgbtRed;
  CMYK.cmykMagenta := 255 - RGB.rgbtGreen;
  CMYK.cmykYellow := 255 - RGB.rgbtBlue;
  if CMYK.cmykCyan < CMYK.cmykMagenta then
    CMYK.cmykK := CMYK.cmykCyan else
    CMYK.cmykK := CMYK.cmykMagenta;
  if CMYK.cmykYellow < CMYK.cmykK then
    CMYK.cmykK := CMYK.cmykYellow;
  if CMYK.cmykK > 0 then begin
    CMYK.cmykCyan := CMYK.cmykCyan - CMYK.cmykK;
    CMYK.cmykMagenta := CMYK.cmykMagenta - CMYK.cmykK;
    CMYK.cmykYellow := CMYK.cmykYellow - CMYK.cmykK;
  end;
end ;


// -----------------------------------------------------------------------------
//
//			CMYKtoRGB
//
// -----------------------------------------------------------------------------
procedure CMYKToRGB(const CMYK:TCMYKQuad;var RGB:TRGBTriple);
begin
   if (Integer(CMYK.cmykCyan) + Integer(CMYK.cmykK)) < 255 then
     RGB.rgbtRed := 255 - (CMYK.cmykCyan + CMYK.cmykK) else
     RGB.rgbtRed := 0;
   if (Integer(CMYK.cmykMagenta) + Integer(CMYK.cmykK)) < 255 then
     RGB.rgbtGreen := 255 - (CMYK.cmykMagenta + CMYK.cmykK) else
     RGB.rgbtGreen := 0;
   if (Integer(CMYK.cmykYellow) + Integer(CMYK.cmykK)) < 255 then
     RGB.rgbtBlue := 255 - (CMYK.cmykYellow + CMYK.cmykK) else
     RGB.rgbtBlue := 0;
end;


procedure CorrectCMYKColors(var CMYK:TCMYKQuad);
var
  MinColor : byte;
begin
  if CMYK.cmykCyan < CMYK.cmykMagenta then
    MinColor := CMYK.cmykCyan else
    MinColor := CMYK.cmykMagenta;
  if CMYK.cmykYellow < MinColor  then
    MinColor := CMYK.cmykYellow;
  if MinColor + CMYK.cmykK > 255 then
    MinColor := 255 - CMYK.cmykK;
  CMYK.cmykCyan := CMYK.cmykCyan - MinColor;
  CMYK.cmykMagenta := CMYK.cmykMagenta - MinColor;
  CMYK.cmykYellow := CMYK.cmykYellow - MinColor;
  CMYK.cmykK := CMYK.cmykK + MinColor;
end;


// -----------------------------------------------------------------------------
//
//			RGBtoHSLDbl
//
// -----------------------------------------------------------------------------
  // RGB, each 0 to 255, to HLS.
  // H = 0 to 359 (corresponding to 0..359 degrees around hexcone)
  // S = 0 (shade of gray) to 255 (pure color)
  // S = 0 (black) to 255 {white)
  //
  // Based on C Code in "Computer Graphics -- Principles and Practice,"
  // Foley et al, 1996, p. 595.  Floating point fractions, 0..1, replaced with
  // integer values, 0..255.
procedure RGBtoHSLDbl (RGB: TRGBTriple; var H, S, L : double);
var
  R,G,B,Delta,Max,
  Min: double;

begin
  R :=  RGB.rgbtRed / 255;
  G :=  RGB.rgbtGreen / 255;
  B :=  RGB.rgbtBlue / 255;
  MinMaxDouble3(R,G,B,Min,Max);
{ Max := MaxDouble2 (R, MaxDouble2 (G, B));
  Min := MinDouble2 (R, MinDouble2 (G, B));}

// calculate luminosity
  L := (Max + Min) / 2;

  if Max = Min then  // it's grey
  begin
    H := 0; // it's actually undefined
    S := 0
  end else begin
    Delta := Max - Min;

// calculate Saturation
    if L < 0.5 then
      S := Delta / (Max + Min)
    else
      S := Delta / (2 - Max - Min);

// calculate Hue
    if R = Max then
      H := (G - B) / Delta
    else
      if G = Max then
        H  := 2 + (B - R) /Delta
      else

        H := 4 + (R - G) / Delta;

    H := H / 6;
    if H < 0 then
      H := H + 1
  end
end;

// -----------------------------------------------------------------------------
//
//			RGBtoHSL
//
// -----------------------------------------------------------------------------
procedure RGBtoHSL(RGB: TRGBTriple; var HSL : THSLTriple);
var
  Hd,
  Sd,
  Ld: double;
begin
  RGBtoHSLDbl (RGB, Hd, sd, ld);
  HSL.hslHue := round (hd * (359-1));
  HSL.hslSaturation := round (sd * 255);
  HSL.hslLightness := round (ld * 255);
end;

  {RGB, each 0 to 255, to HSV.
   H = 0 to 359 (corresponding to 0..359 degrees around hexcone)
   S = 0 (shade of gray) to 255 (pure color)
   V = 0 (black) to 255 {white)

   Based on C Code in "Computer Graphics -- Principles and Practice,"
   Foley et al, 1996, p. 592.  Floating point fractions, 0..1, replaced with
   integer values, 0..255.
  }

  procedure RGBtoHSV (const RGB:    TRGBTriple;  {r, g and b IN [0..255]}
                      var   HSV:  THSVTriple);    {h IN 0..359; s,v IN 0..255}
    var
      Delta   :  INTEGER;
      MinValue:  INTEGER;
  begin
    with RGB DO
      MinMaxInt3(rgbtRed, rgbtGreen, rgbtBlue, MinValue, HSV.hsvValue);
    Delta := HSV.hsvValue - MinValue;

    {Calculate saturation:  saturation is 0 if r, g and b are all 0}
    if   HSV.hsvValue = 0
    then HSV.hsvSaturation := 0
    else HSV.hsvSaturation := (255 * Delta) div HSV.hsvValue;

    if   HSV.hsvSaturation = 0
    then HSV.hsvHue := 0   {Achromatic:  When s = 0, h is undefined but assigned the value 0}
    else begin    {Chromatic}

      with RGB DO
      begin
        if   rgbtRed = HSV.hsvValue
        then HSV.hsvHue := (60*(rgbtGreen-rgbtBlue)) div Delta            {degrees -- between yellow and magenta}
        else
          if   rgbtGreen = HSV.hsvValue
          then HSV.hsvHue := 120 + (60*(rgbtBlue-rgbtRed)) div Delta      {degrees -- between cyan and yellow}
          else
            if  rgbtBlue = HSV.hsvValue
            then HSV.hsvHue := 240 + (60*(rgbtRed-rgbtGreen)) div Delta;  {degrees -- between magenta and cyan}
      end;

      if   HSV.hsvHue < 0
      then HSV.hsvHue := HSV.hsvHue + 360;

    end
  end {RGBtoHSV};

// -----------------------------------------------------------------------------
//
//			YCbCrToRGB
//
// -----------------------------------------------------------------------------
procedure YCbCrToRGB(YCbCr:TYCbCrTriple; var RGB:  TRGBTRiple);
const C=256;
      c11:real= 0.0054980*C;
      c12:real= 0.0000000*C;
      c13:real= 0.0051681*C;
      c21:real= 0.0054980*C;
      c22:real=-0.0015446*C;
      c23:real=-0.0026325*C;
      c31:real= 0.0054980*C;
      c32:real= 0.0079533*C;
      c33:real= 0.0000000*C;
var   r,g,b:Integer;
begin

  r:=round(c11*YCbCr.ycbcrY +c12*(YCbCr.ycbcrCB-156) +c13*(YCbCr.ycbcrCR-137));
  g:=round(c21*YCbCr.ycbcrY +c22*(YCbCr.ycbcrCB-156) +c23*(YCbCr.ycbcrCR-137));
  b:=round(c31*YCbCr.ycbcrY +c32*(YCbCr.ycbcrCB-156) +c33*(YCbCr.ycbcrCR-137));
  if r<0   then r:=0;
  if g<0   then g:=0;
  if b<0   then b:=0;
  if r>255 then r:=255;
  if g>255 then g:=255;
  if b>255 then b:=255;
  with RGB do
  begin
    rgbtRed:=r;
    rgbtGreen:=g;
    rgbtBlue:=b;
  end;
end;

// -----------------------------------------------------------------------------
//
//			RGBAtoBGRA
//
// -----------------------------------------------------------------------------
procedure RGBAtoBGRA(RGBA:TRGBAQuad; var BGRA:TBGRAQuad);
begin
    BGRA.bgrBlue := RGBA.rgbBlue;
    BGRA.bgrRed := RGBA.rgbRed;
    BGRA.bgrGreen := RGBA.rgbGreen;
    BGRA.bgrAlpha := RGBA.rgbAlpha;
end;

// -----------------------------------------------------------------------------
//
//			RGBAtoBGRA_16
//
// -----------------------------------------------------------------------------
procedure RGBAtoBGRA_16(RGBA:TRGBAQuad; var BGRA:TBGRAQuad);
begin
    BGRA.bgrBlue := RGBA.rgbRed shr 8;
    BGRA.bgrRed := RGBA.rgbBlue shr 8;
    BGRA.bgrGreen := RGBA.rgbGreen shr 8;
    BGRA.bgrAlpha := RGBA.rgbAlpha shr 8;
end;


// -----------------------------------------------------------------------------
//
//			CIELABtoBGR
//
// -----------------------------------------------------------------------------
procedure CIELABtoBGR(CIELABTriple:TCIELABTriple; var BGRTriple:TBGRTriple);
var
  L, a, b,
  X, Y, Z, T,  YYn3: Double;

begin
    L := CIELABTriple.L / 2.56;
    A := CIELABTriple.A;
    B := CIELABTriple.B;

    YYn3 := (L + 16) / 116;
    if L < 7.9996 then
    begin
      Y := L / 903.3;
      X := A / 3893.5 + Y;
      Z := Y - b / 1557.4;
    end
    else
    begin
         T := YYn3 + a / 500;
	 X := T * T * T;
	 Y := YYn3 * YYn3 * YYn3;
	 T := YYn3 - b / 200;
	 Z := T * T * T;
    end;
    BGRTriple.bgrRed:= TrimInt(0,255,Round(255 * ( 2.998 * X - 1.458 * Y - 0.541 * Z)));
    BGRTriple.bgrGreen:= TrimInt(0,255,Round(255 * (-0.952 * X + 1.893 * Y + 0.059 * Z)));
    BGRTriple.bgrBlue:=TrimInt(0,255,Round(255 * ( 0.099 * X - 0.198 * Y + 1.099 * Z)));

end;


// -----------------------------------------------------------------------------
//
//			CalculateGamma
//
// -----------------------------------------------------------------------------
function CalculateGamma(Gamma:Double;Value:Integer):Double;
begin
 if Value >255 then Value:=255;
 if Value >0 then
 result:=255* exp (Gamma * ln(Value/255))+0.5
 else result:=0;
end;

// -----------------------------------------------------------------------------
//
//			IntGamma
//
// -----------------------------------------------------------------------------
function IntGamma(Gamma:Double;Value:Integer):Integer;
begin
  result:=TrimReal(0,255,CalculateGamma(Gamma,Value));
end;


// -----------------------------------------------------------------------------
//
//			GetColorDepth
//
// -----------------------------------------------------------------------------
  // Adapted from Tommy Andersen's ColorDepth post to
  // comp.lang.pascal.delphi.components.misc, 5 Oct 1997.
  //
  // According to Tim Robert's post "Bitmaps with > 256 Colors" to
  // compl.lang.pascal.delphi.misc, 17 Apr 1997, "in a windows bitmap,
  // one of either the bits-per-plane or the number of planes must be 1.  In
  // fact, the planar form is only supported at 4 bits per pixel.  A 24-bit
  // true color bitmap must have bits-per-pixel set to 24 and planes set to 1.
function GetColorDepth:  INTEGER;
var
      DeviceContext:  HxDC;
begin
    // Get the screen's DC since memory DCs are not reliable
    DeviceContext := xGetDC(0);

    TRY
      RESULT := 1 SHL (xGetDeviceCaps(DeviceContext, xPLANES) *
                       xGetDeviceCaps(DeviceContext, xBITSPIXEL))
    FINALLY
      // Give back the screen DC
      xReleaseDC(0, DeviceContext)
    end;
end;


// -----------------------------------------------------------------------------
//
//			GetPixelDepth
//
// -----------------------------------------------------------------------------
function  GetPixelDepth:  INTEGER;
    var
      DeviceContext:  HxDC;
begin
    // Get the screen's DC since memory DCs are not reliable
    DeviceContext := xGetDC(0);

    TRY
      RESULT := xGetDeviceCaps(DeviceContext, xBITSPIXEL)
    FINALLY
      // Give back the screen DC
      xReleaseDC(0, DeviceContext)
    end;
end;

// -----------------------------------------------------------------------------
//
//			IsPaletteDevice
//
// -----------------------------------------------------------------------------
  //  Adapted from Joe C. Hecht's BitTXBitmapAsDIB post to
  //  borland.public.delphi.winapi, 12 Oct 1997.
function IsPaletteDevice:  BOOLEAN;
    var
      DeviceContext:  HxDC;
begin
    // Get the screen's DC since memory DCs are not reliable
    DeviceContext := xGetDC(0);

    TRY
      RESULT := xGetDeviceCaps(DeviceContext, xRASTERCAPS) AND xRC_PALETTE = xRC_PALETTE
    FINALLY
      // Give back the screen DC
      xReleaseDC(0, DeviceContext)
    end
end;

{==  Bitmaps  =======================================================}

// -----------------------------------------------------------------------------
//
//			PrinBitmap
//
// -----------------------------------------------------------------------------
  {Based on posting to borland.public.delphi.winapi by Rodney E Geraghty, 8/8/97.
   Used to print bitmap on any Windows printer.}
procedure PrintBitmap(Canvas:  TXCanvas; DestRect:  TXRect;  Bitmap:  TXBitmap);
var
      BitmapHeader:  pXBitmapInfo;
      BitmapImage :  POINTER;
      HeaderSize  :  xDWOrd;
      ImageSize   :  xDWord;
begin
    xGetDIBSizes(Bitmap.Handle, HeaderSize, ImageSize);
    GetMem(BitmapHeader, HeaderSize);
    GetMem(BitmapImage,  ImageSize);
    TRY
      xGetDIB(Bitmap.Handle, Bitmap.Palette, BitmapHeader^, BitmapImage^);
      xStretchDIBits(Canvas.Handle,
                    DestRect.Left, DestRect.Top,     {Destination Origin}
                    DestRect.Right  - DestRect.Left, {Destination Width}
                    DestRect.Bottom - DestRect.Top,  {Destination Height}
                    0, 0,                            {Source Origin}
                    Bitmap.Width, Bitmap.Height,     {Source Width & Height}
                    BitmapImage,
                    TXBitmapInfo(BitmapHeader^),
                    xDIB_RGB_COLORS,
                    xSRCCOPY)
    FINALLY
      FreeMem(BitmapHeader);
      FreeMem(BitmapImage)
    end
end;


// -----------------------------------------------------------------------------
//
//			GeTXBitmapDimensionsString
//
// -----------------------------------------------------------------------------
function GetBitmapDimensionsString(const Bitmap:  TXBitmap):  STRING;
begin
    RESULT := xIntToStr(Bitmap.Width)  + ' by ' +
              xIntToStr(Bitmap.Height) + ' pixels by ' +
              GetPixelformatString(Bitmap.Pixelformat) + ' color';
end;

// -----------------------------------------------------------------------------
//
//			GetColorPlaneString
//
// -----------------------------------------------------------------------------
function GetColorPlaneString(const ColorPlane:  TColorPlane):  STRING;
begin
    CASE ColorPlane OF
      cpRGB:           RESULT := 'RGB Composite';

      cpRed:           RESULT := 'Red';
      cpGreen:         RESULT := 'Green';
      cpBlue:          RESULT := 'Blue';

      cpHueHSV:        RESULT := 'Hue (HSV)';
      cpSaturationHSV: RESULT := 'Saturation (HSV)';
      cpValue:         RESULT := 'Value (HSV)';

      cpCyan:          RESULT := 'Cyan';
      cpMagenta:       RESULT := 'Magenta';
      cpYellow:        RESULT := 'Yellow';
      cpBlack:         RESULT := 'Black (CMYK)';

      cpIntensity:     RESULT := 'RGB Intensity';
    end

end;


// -----------------------------------------------------------------------------
//
//			GetPixelFormatString
//
// -----------------------------------------------------------------------------
function GetPixelformatString(const Pixelformat:  TXPixelformat):  STRING;
    var
      format:  STRING;
begin
    CASE Pixelformat OF
      xpfDevice:  format := 'Device';
      xpf1bit:    format := '1 bit';
      xpf4bit:    format := '4 bit';
      xpf8bit:    format := '8 bit';
      xpf15bit:   format := '15 bit';
      xpf16bit:   format := '16 bit';
      xpf24bit:   format := '24 bit';
      xpf32bit:   format := '32 bit'
      else
        format := 'Unknown';
    end;
    RESULT := format;
end;


// -----------------------------------------------------------------------------
//                       *******************
//                       Rotate Filter Matrix
//                       *******************
// -----------------------------------------------------------------------------
//
//     Bassic parameters:
//               Filter  : TGrphicFilter
//               Steps   : Steps
// -----------------------------------------------------------------------------

procedure RotateFilterMatrix(var Filter:TGraphicFilter;Steps:Integer);
var hm:TEffectMatrix;
    k,i,j:Integer;

procedure ActualizeMatrix;
var i,j:Integer;
begin
   for i:=0 to 6 do for j:=0 to 6 do
   Filter.Matrix[i,j]:=hm[i,j];
end;

procedure Rotate3x3;
  begin
  with Filter do begin
    hm[3,3]:=Matrix[3,3];
    hm[2,2]:=Matrix[3,2];
    hm[3,2]:=Matrix[4,2];
    hm[4,2]:=Matrix[4,3];
    hm[4,3]:=Matrix[4,4];
    hm[4,4]:=Matrix[3,4];
    hm[3,4]:=Matrix[2,4];
    hm[2,4]:=Matrix[2,3];
    hm[2,3]:=Matrix[2,2];
   end;
   ActualizeMatrix;
  end;

procedure Rotate5x5;
var  i,j:Integer;
  begin
   with Filter do begin
    for i:=1 to 3 do hm[i,1]:=Matrix[i+2,1];
    hm[4,1]:=Matrix[5,2];
    for j:=1 to 3 do hm[5,j]:=MAtrix[5,j+2];
    hm[5,4]:=Matrix[4,5];
    for i:=5 downto 3 do hm[i,5]:=Matrix[i-2,5];
    hm[2,5]:=Matrix[1,4];
    for j:=5 downto 3 do hm[1,j]:=MAtrix[1,j-2];
    hm[1,2]:=Matrix[2,1];
   end;
   ActualizeMatrix;
  end;

procedure Rotate7x7;
var  i,j:Integer;
  begin
   with Filter do begin
    for i:=0 to 3 do hm[i,0]:=Matrix[i+3,0];
    hm[4,0]:=Matrix[6,1];
    hm[5,0]:=Matrix[6,2];
    for j:=0 to 3 do hm[6,j]:=MAtrix[6,j+3];
    hm[6,4]:=Matrix[5,6];
    hm[6,5]:=Matrix[4,6];

    for i:=6 downto 3 do hm[i,6]:=Matrix[i-3,6];
    hm[2,6]:=Matrix[0,5];
    hm[1,6]:=Matrix[0,4];
    for j:=6 downto 3 do hm[0,j]:=MAtrix[0,j-3];
    hm[0,2]:=Matrix[1,0];
    hm[0,1]:=Matrix[2,0];
   end;
   ActualizeMatrix;
  end;

begin
  for i:=0 to 6 do for j:=0 to 6 do hm[i,j]:=  Filter.Matrix[i,j];
  if Steps=0 then  ActualizeMatrix else
  for k:=0 to Steps-1 do with Filter do
   begin
   Rotate3x3;
   case MatrixSize of
   mx5:Rotate5x5;
   mx7:Begin
        Rotate5x5;
        Rotate7x7;
       end;
   end;
  end;
end;



  procedure QSort(var X : TIntVector; Lbound, Ubound : Integer);
  { Quick sort in ascending order - Adapted from Borland's BP7 demo }
    procedure Sort(L, R : Integer);
    var
      I, J : Integer;
      U, V : Integer;
    begin
      I := L;
      J := R;
      U := X[(L + R) div 2];
      repeat
        while X[I] < U do I := I + 1;
        while U < X[J] do J := J - 1;
        if I <= J then
          begin
            V := X[I]; X[I] := X[J]; X[J] := V;
            I := I + 1; J := J - 1;
          end;
      until I > J;
      if L < J then Sort(L, J);
      if I < R then Sort(I, R);
    end;
begin
  Sort(LBound,UBound);
end;

procedure MatrixQSort(var X : TIntXYVector;Size:Integer);
    procedure Sort(L, R : Integer);
    var
      I, J : Integer;
      X1,Y1: Integer;
      U, V : Integer;
    begin
      I := L;
      J := R;
      U := X[(L + R) div 2].Int;
      repeat
        while X[I].Int < U do I := I + 1;
        while U < X[J].Int do J := J - 1;
        if I <= J then
          begin
            V := X[I].Int;
            X1:= X[I].X;
            Y1:= X[I].Y;

            X[I].Int := X[J].Int;
            X[I].X := X[J].X;
            X[I].Y := X[J].Y;

            X[J].Int := V;
            X[J].X := X1;
            X[J].Y := Y1;

            I := I + 1; J := J - 1;
          end;
      until I > J;
      if L < J then Sort(L, J);
      if I < R then Sort(I, R);
    end;
begin
  Sort(0,Size);
end;

function GetStatOfMatrix(Matrix:TEffectMatrix;SFType:TSTaticFilterType;Diameter:Integer;var Row,Col:Integer):Integer;
var i,x1,y1:Integer;
    x:    TIntXYVector;

begin
  i:=0;
  for Y1:=-Diameter to Diameter do
  for X1:=-Diameter to Diameter do
  begin
    X[i].Int:=Matrix[Y1+3,X1+3];
    X[i].X:=X1;
    X[i].Y:=Y1;
    inc(i);
  end;
  MatrixQSort(X,Diameter*2);
  case SFType of
  sfMin: begin
              result:=x[0].Int;
              Row:=x[0].Y;
              Col:=x[0].X;
            end;
  sfMax: begin
              result:= x[Diameter*2].Int;
              Row:=x[Diameter*2].Y;
              Col:=x[Diameter*2].X;
            end;
  else     begin
              result:= x[Diameter].Int;
              Row:=x[Diameter].Y;
              Col:=x[Diameter].X;
            end;
  end;
end;

procedure LoadLinearFilterFromFile(FileName:PChar;var Filter:TGraphicFilter);stdcall;
var Stream:TFileStream;
    header: Array [0..5] of Char;
begin
 Stream:=TFileStream.Create(FileNAme,fmOpenRead);
 try
   ZeroMemory(PChar(@header),sizeof(header));
   Stream.Read(header,4);
   if header<>'fxlf' then raise EGraphicEffects.Create('Not a valid linear filter file') else
   begin
      Stream.Read(Filter,SizeOf(TGraphicFilter));

   end;
 finally
   Stream.Free;
 end;
end;

procedure SaveLinearFilterToFile(FileName:PChar;const Filter:TGraphicFilter);stdcall;
var stream:TFileStream;
 const c ='Created by pView (c) 1999 by A.Moser';
begin
 Stream:=TFileStream.Create(FileNAme,fmCreate);
 try
   Stream.Write('fxlf',4);
   Stream.Write(Filter,SizeOf(TGraphicFilter));
   Stream.Write(c,Length(c));
 finally
   Stream.Free;
 end;

end;

procedure LoadMultiPassFilterFromFile(FileName:PChar;var Filter:TMultiPassGraphicFilter);stdcall;
var Stream:TFileStream;
    header: Array [0..5] of Char;
begin
 Stream:=TFileStream.Create(FileNAme,fmOpenRead);
 try
   ZeroMemory(PChar(@header),sizeof(header));
   Stream.Read(header,4);
   if header<>'fxmf' then raise EGraphicEffects.Create('Not a valid linear filter file') else
   begin
      Stream.Read(Filter,SizeOf(TMultiPassGraphicFilter));

   end;
 finally
   Stream.Free;
 end;
end;

procedure SaveMultipassFilterToFile(FileName:PChar;const Filter:TMultiPassGraphicFilter);stdcall;
var stream:TFileStream;
 const c ='Created by pView (c) 1999 by A.Moser';
begin
 Stream:=TFileStream.Create(FileNAme,fmCreate);
 try
   Stream.Write('fxmf',4);
   Stream.Write(Filter,SizeOf(TMultiPassGraphicFilter));
   Stream.Write(c,Length(c));
 finally
   Stream.Free;
 end;

end;


end.
